# -*- coding: utf-8 -*-
import logging
import sys

__copyright__ = u"Copyright (c), This file is part of the AiiDA platform. For further information please visit http://www.aiida.net/. All rights reserved."
__license__ = "MIT license, see LICENSE.txt file."
__version__ = "0.7.0.1"
__authors__ = "The AiiDA team."

aiidalogger = logging.getLogger("aiida")
# aiidalogger.addHandler(logging.StreamHandler(sys.stderr))
#FORMAT = '[%(name)s@%(levelname)s] %(message)s'
#logging.basicConfig(format=FORMAT)
